use std::{
    ops::{Add, Sub},
    process::exit,
    usize,
};

use indicatif::ProgressStyle;

fn main() {
    let args: Vec<String> = std::env::args().collect();
    let program = std::fs::read_to_string(args.get(1).unwrap()).unwrap();

    let mut machine = Machine::new(program);
    if args.get(2).is_some() {
        loop {
            let mut input = String::new();
            std::io::stdin().read_line(&mut input).unwrap();
            for s in input.trim().chars() {
                match s {
                    's' => {
                        let state = machine.step();
                        if state == MachineState::EOI {
                            println!("Reached end of program printing end state");
                            machine.instruction_pointer -= 1;
                            machine.dbg();
                            machine.mem();
                            machine.code();
                            exit(0);
                        }
                    }
                    'd' => {
                        machine.dbg();
                    }
                    'm' => {
                        machine.mem();
                    }
                    'i' => {
                        machine.code();
                    }
                    'w' => {
                        machine.write();
                    }
                    'e' => {
                        machine.execute();
                        machine.instruction_pointer -= 1;
                    }
                    'r' => machine.run_for(),
                    _ => {
                        println!("Option unrecognized")
                    }
                }
            }
        }
    } else {
        machine.execute();
    }
}

#[derive(Debug)]
struct Machine {
    instruction_pointer: usize,
    data_pointer: usize,
    data: Vec<Cell>,
    instructions: Vec<Instruction>,
}


#[derive(Debug, Default, Clone, Copy, PartialEq)]
struct Cell {
    value: u8,
}

impl Machine {
    fn new(program: String) -> Self {
        Machine {
            instruction_pointer: 0,
            data_pointer: 0,
            data: vec![Cell { value: 0 }],
            instructions: parse(program),
        }
    }

    fn execute(&mut self) {
        loop {
            match self.step() {
                MachineState::Ok => continue,
                MachineState::EOI => break,
            }
        }
    }

    fn step(&mut self) -> MachineState {
        let instruction = self.instructions.get(self.instruction_pointer).unwrap();
        match instruction {
            Instruction::BINC => self.binc(),
            Instruction::BDEC => self.bdec(),
            Instruction::PINC => self.pinc(),
            Instruction::PDEC => self.pdec(),
            Instruction::JIZ(target) => self.jiz(*target),
            Instruction::JINZ(target) => self.jinz(*target),
            Instruction::OUT => self.out(),
            Instruction::IN => self.inp(),
            Instruction::DBG => self.dbg(),
            Instruction::MEM => self.mem(),
        }
        self.instruction_pointer += 1;
        if self.instruction_pointer + 1 > self.instructions.len() {
            MachineState::EOI
        } else {
            MachineState::Ok
        }
    }

    fn binc(&mut self) {
        let cell = *self
            .data
            .get_mut(self.data_pointer)
            .unwrap_or(&mut Cell { value: 0 })
            + Cell { value: 1 };
        self.data[self.data_pointer] = cell;
    }

    fn bdec(&mut self) {
        let cell = *self
            .data
            .get_mut(self.data_pointer)
            .unwrap_or(&mut Cell { value: 0 })
            - Cell { value: 1 };
        self.data[self.data_pointer] = cell;
    }

    fn pinc(&mut self) {
        self.data_pointer += 1;
        if self.data.get(self.data_pointer).is_none() {
            self.data.push(Cell { value: 0 });
        }
    }

    fn pdec(&mut self) {
        if self.data_pointer != 0 {
            self.data_pointer -= 1;
        }
    }

    fn jiz(&mut self, target: usize) {
        if self
            .data
            .get(self.data_pointer)
            .unwrap_or(&Cell { value: 0 })
            .value
            == 0
        {
            self.instruction_pointer = target;
        }
    }

    fn jinz(&mut self, target: usize) {
        if self
            .data
            .get(self.data_pointer)
            .unwrap_or(&Cell { value: 0 })
            .value
            != 0
        {
            self.instruction_pointer = target;
        }
    }

    fn out(&self) {
        let value = self
            .data
            .get(self.data_pointer)
            .unwrap_or(&Cell { value: 0 })
            .value;
        let c: char = value.into();
        print!("{c}");
    }

    fn inp(&mut self) {
        let mut input = String::new();
        std::io::stdin().read_line(&mut input).unwrap();

        let value = *input.into_bytes().first().unwrap();

        let cell = Cell { value };

        self.data[self.data_pointer] = cell;
    }

    fn dbg(&self) {
        println!(
            "Instruction pointer: {} \nCurrent Instruction: {:?}\nData pointer: {}",
            self.instruction_pointer,
            self.instructions[self.instruction_pointer],
            self.data_pointer
        );
    }

    fn mem(&self) {
        println!("Memory: {:?}", self.data);
    }

    fn code(&self) {
        println!("Instructions: {:?}", self.instructions);
    }

    fn write(&mut self) {
        let mut input = String::new();
        std::io::stdin().read_line(&mut input).unwrap();

        let addres: usize = input.trim().parse().unwrap();

        input = String::new();
        std::io::stdin().read_line(&mut input).unwrap();
        let value: u8 = input.trim().parse().unwrap();

        let cell = Cell { value };

        if self.data.len() <= self.data_pointer {
            self.data.insert(addres, cell);
        } else {
            self.data[addres] = cell;
        }

        println!("Wrote {value} to {addres}");
    }

    fn run_for(&mut self) {
        let mut input = String::new();
        std::io::stdin().read_line(&mut input).unwrap();

        let repetitions: usize = input.trim().parse().unwrap();

        let bar = indicatif::ProgressBar::new(repetitions as u64);
        bar.set_style(
            ProgressStyle::with_template(
                "{bar:40} | {human_pos}/{human_len} | {eta_precise} | {elapsed_precise}",
            )
            .unwrap(),
        );
        for _ in 0..repetitions {
            bar.inc(1);
            match self.step() {
                MachineState::Ok => continue,
                MachineState::EOI => {
                    self.instruction_pointer -= 1;
                    break;
                }
            }
        }
        bar.finish();
    }
}

impl Add for Cell {
    type Output = Cell;

    fn add(self, rhs: Self) -> Self::Output {
        Cell {
            value: self.value + rhs.value,
        }
    }
}

impl Sub for Cell {
    type Output = Cell;

    fn sub(self, rhs: Self) -> Self::Output {
        if self.value == 0 {
            Cell { value: 0 }
        } else {
            Cell {
                value: self.value - rhs.value,
            }
        }
    }
}
#[allow(clippy::upper_case_acronyms)]
#[derive(PartialEq, Eq, Debug)]
enum Instruction {
    PINC,        // >
    PDEC,        // <
    BINC,        // +
    BDEC,        // -
    OUT,         // .
    IN,          // ,
    JIZ(usize),  // [
    JINZ(usize), // ]
    DBG,         // ? (Special instruction not standart)
    MEM,         // ! (Special instruction not standart)
}

#[derive(Debug)]
enum ParseError {
    NoTargetForJIZ(usize),
    NoTargetForJINZ(usize),
}
#[derive(Debug, PartialEq)]
#[allow(clippy::upper_case_acronyms)]
enum MachineState {
    Ok,
    EOI, // End of Instructions
}
fn parse(program: String) -> Vec<Instruction> {
    let mut instruction_vec: Vec<Instruction> = vec![];

    for (i, c) in program.chars().enumerate() {
        if let Some(instruction) =
            match_single_char_populate(&c, &program, i, instruction_vec.len())
        {
            instruction_vec.push(instruction);
        }
    }

    instruction_vec
}

fn match_single_char_populate(
    c: &char,
    program: &String,
    index: usize,
    instruction_index: usize,
) -> Option<Instruction> {
    match c {
        '>' => Some(Instruction::PINC),
        '<' => Some(Instruction::PDEC),
        '+' => Some(Instruction::BINC),
        '-' => Some(Instruction::BDEC),
        '.' => Some(Instruction::OUT),
        ',' => Some(Instruction::IN),
        '[' => Some(Instruction::JIZ(
            locate_jmp_target_jiz(program, index, instruction_index).unwrap(),
        )),
        ']' => Some(Instruction::JINZ(
            locate_jmp_target_jinz(program, index, instruction_index).unwrap(),
        )),
        '?' => Some(Instruction::DBG),
        '!' => Some(Instruction::MEM),
        _ => None,
    }
}

fn match_single_char(c: &char) -> Option<Instruction> {
    match c {
        '>' => Some(Instruction::PINC),
        '<' => Some(Instruction::PDEC),
        '+' => Some(Instruction::BINC),
        '-' => Some(Instruction::BDEC),
        '.' => Some(Instruction::OUT),
        ',' => Some(Instruction::IN),
        '[' => Some(Instruction::JIZ(0)),
        ']' => Some(Instruction::JINZ(0)),
        '?' => Some(Instruction::DBG),
        '!' => Some(Instruction::MEM),
        _ => None,
    }
}

fn locate_jmp_target_jiz(
    program: &String,
    index: usize,
    instruction_index: usize,
) -> Result<usize, ParseError> {
    let program_vector: Vec<char> = program.chars().collect();
    let mut data_pointer = index + 1;
    let mut searched_jinz = 1;
    let mut instruction_index = instruction_index;
    instruction_index = instruction_index.saturating_sub(1);

    while searched_jinz != 0 {
        if data_pointer == program_vector.len() {
            return Err(ParseError::NoTargetForJIZ(index));
        }

        if let Some(instruction) = match_single_char(program_vector.get(data_pointer).unwrap()) {
            match instruction {
                Instruction::JIZ(_) => searched_jinz += 1,
                Instruction::JINZ(_) => searched_jinz -= 1,
                _ => (),
            }
            instruction_index += 1;
        }
        data_pointer += 1;
    }

    Ok(instruction_index)
}

fn locate_jmp_target_jinz(
    program: &String,
    index: usize,
    instruction_index: usize,
) -> Result<usize, ParseError> {
    let program_vector: Vec<char> = program.chars().collect();
    let mut data_pointer = index - 1;
    let mut searched_jiz = 1;
    let mut instruction_index = instruction_index - 1;

    while searched_jiz != 0 {
        if let Some(instruction) = match_single_char(program_vector.get(data_pointer).unwrap()) {
            match instruction {
                Instruction::JINZ(_) => searched_jiz += 1,
                Instruction::JIZ(_) => searched_jiz -= 1,
                _ => (),
            }
            instruction_index -= 1;
        } else if data_pointer == 0 {
            return Err(ParseError::NoTargetForJINZ(index));
        }
        data_pointer -= 1;
    }

    Ok(instruction_index + 1)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn locate_jmp_target_for_jiz_single_jmp() {
        let program = String::from("[++>]-");
        let index = 0;

        let return_value = locate_jmp_target_jiz(&program, index, index).unwrap();

        assert_eq!(return_value, 4);
    }

    #[test]
    fn locate_jmp_target_for_jiz_multi_jmp() {
        let program = String::from("+[>+[<]-]+");
        let index = 1;

        let return_value = locate_jmp_target_jiz(&program, index, index).unwrap();

        assert_eq!(return_value, 7);
    }

    #[test]
    fn locate_jmp_target_for_jinz_single_jmp() {
        let program = String::from("+[>>+]");
        let index = 5;

        let return_value = locate_jmp_target_jinz(&program, index, index).unwrap();

        assert_eq!(return_value, 1);
    }

    #[test]
    fn locate_jmp_target_for_jinz_multi_jmp() {
        let program = String::from("+[>[<]>+]");
        let index = 8;

        let return_value = locate_jmp_target_jinz(&program, index, index).unwrap();

        assert_eq!(return_value, 1);
    }

    #[test]
    fn parse_program_no_jmp() {
        let program = String::from("+><-.>,");

        let program_vec = parse(program);
        let correct_output = vec![
            Instruction::BINC,
            Instruction::PINC,
            Instruction::PDEC,
            Instruction::BDEC,
            Instruction::OUT,
            Instruction::PINC,
            Instruction::IN,
        ];

        for (index, instruction) in program_vec.iter().enumerate() {
            assert_eq!(instruction, correct_output.get(index).unwrap());
        }
    }

    #[test]
    fn parse_program_single_jmp() {
        let program = String::from("+>[<]-.>,");

        let program_vec = parse(program);
        let correct_output = vec![
            Instruction::BINC,
            Instruction::PINC,
            Instruction::JIZ(3),
            Instruction::PDEC,
            Instruction::JINZ(2),
            Instruction::BDEC,
            Instruction::OUT,
            Instruction::PINC,
            Instruction::IN,
        ];

        for (index, instruction) in program_vec.iter().enumerate() {
            dbg!(&index);
            assert_eq!(instruction, correct_output.get(index).unwrap());
        }
    }

    #[test]
    fn add_values() {
        let program = String::from("+++>++++[<+>-]");
        let mut machine = Machine::new(program);

        machine.execute();
        let cell0 = Cell { value: 7 };
        assert_eq!(machine.data[0], cell0);
    }
}
